//	74.442	Advanced Design & Analysis of Algorithms
//	Spring 1998
//	Dr. Bill Kocay
//
//	Hamish Carr
//	Project:	Toroidal Surface Geometries
//
//

#pragma once

//	We will express points on a tile in terms of the four corners:
//		TL	TR		(i.e.)		theta0, phi0	theta1, phi0
//		BL	BR					theta0, phi1	theta1, phi1
//	
//	We *could* specify any point on the tile as (th, ph), th, ph in [0, 1]
//		
//	But, because we have to deal with screen points that are neither vertical or horizontal, this leads to some
//	very messy calculations to find the exact point on the screen. 
//	
//	Since we know that the endpoints of the segment are to be found on sides of the tile, we know that we will be
//	calculating x*corner1 + (1-x)*corner2, with the other 2 corners not contributing anything.
//	The simplest way of doing this is to keep weights for all 4 corners, 2 of which will be zero for any point
//	the other 2 of which will sum to 1.0

class Segment																//	a segment of a line
	{
	public:
	float inTL, inTR, inBL, inBR;											//	relative components of 4 corners for entry point
	float outTL, outTR, outBL, outBR;										//	relative components of 4 corners for exit point
	long thetaIncrement, phiIncrement;										//	increments for moving to tile under next segment
	}; // end of Segment class

//	Since we will use the same segment multiple times for different tiles, we need the linked list for a given tile to be independent
//	of the segment.  The easiest way is:

class SegmentNode															//	a node pointing to a segment
	{
	public:
	Segment *theSegment;													//	the segment proper
	SegmentNode *nextSegment;												//	pointer to the next segment node
	}; // end of SegmentNode class
	
class Line																	//	an array of segments
	{
	public:
	Segment *theSegments;													//	an array of the segments
	long nSegments;															//	how many segments in the line
	long theta, phi;														//	position of a point on the line
	long dTheta, dPhi;														//	distance to next tile (as a unit)
	bool isSelected;														//	whether the line is currently selected
	}; // end of Line class
	
typedef Line *LinePointer;													//	a pointer to a line