//	74.442	Advanced Design & Analysis of Algorithms
//	Spring 1998
//	Dr. Bill Kocay
//
//	Hamish Carr
//	Project:	Toroidal Surface Geometries
//
//

#pragma once

class Torus;																//	forward declaration of torus class

#include "Tile.h"
#include "GCDLCM.h"
#include "GlobalRoutines.h"
#include "UIMessageList.h"
#include "CharBuffer.h"

#define selectionTool 	0L													//	selects a point
#define pointTool 		1L													//	creates a new point
#define lineTool 		2L													//	creates a new line
#define handTool		3L													//	moves the entire object
#define connectTool 	4L													//	connects 2 or more selected points
#define intersectTool	5L													//	intersects 2 or more selected lines
#define rotateUpTool	6L													//	rotates outside face up round tube
#define rotateDownTool	9L													//	rotates outside face down round tube
#define rotateLeftTool	7L													//	rotates points left round torus
#define rotateRightTool	10L													//	rotates points right round torus

#define maxNPoints 20														//	don't allow more than 20 points
#define maxNLines 10														//	or 10 lines
	
#define torusTitle 		"&Torus"											//	set constants for files
#define tubeTitle 		"Tube Radius:"
#define ringTitle 		"Ring Radius:"
#define nPhiTitle 		"# Phi Divisions:"
#define nThetaTitle 	"# Theta Divisions"
#define nPointsTitle 	"# Points"
#define nLinesTitle 	"# Lines"


class Torus																	//	a class representing the torus proper
	{
	public:
	float radius, ringRadius;												//	radius of the tube of the torus & the ring
	long nPhiDivisions, nThetaDivisions;									//	how many divisions in each direction
	long nTiles;															//	how many tiles in total													
	long nPoints;															//	how many points there are shown
	long nLines;															//	how many lines there are shown
	long nPointsSelected;													//	number of currently selected points
	long nLinesSelected;													//	number of currently selected lines
	
	Vector *theGridVectors;													//	the grid points (in vector format)
	BPoint *theGridPoints;													//	the grid points (after projection)

	Tile *theTiles;															//	array with all the tiles
	TilePointer *sortedTiles;												//	array of pointers to the tiles in sorted order

	TilePointer thePoints[maxNPoints];										//	array holding all the points
	TilePointer selectedPoints[maxNPoints];									//	array holding all the currently selected points
	
	Line theLines[maxNLines];												//	array holding all the lines
	LinePointer selectedLines[maxNLines];									//	array holding all the currently selected lines
	
	BPoint screenCentre;													//	the centre of the viewable area
	float viewDistance;														//	the distance from viewpoint to viewplane
	float viewScale;														//	the scale of the view

	Vector torusOrigin;														//	the origin of the torus
	Basis torusBasis;														//	the basis for the torus
	bool insideTorus;														//	flag for whether viewpoint is inside or outside
	long theCurrentTool;													//	the tool currently in use
	
	char *name;																//	file name for saving the torus
	BMessage *itsLocation;													//	stored message showing where it is on disk
	
	Torus(float RingRadius = 2000, float Radius = 1000, long NPhiDivisions = 24, long NThetaDivisions = 120);	
																			//	basic constructor
	Torus(char *torusText, BMessage *newLocation = NULL);					//	string constructor
	~Torus();																//	short & sweet destructor

	void ConstructTiles(float RingRadius, float Radius, long NPhiDivisions, long NThetaDivisions);
																			//	constructs the tiles for the torus
	void SetRenderingParms(BPoint ScreenCentre, float ViewDistance, float ViewScale);
																			//	sets the variables for rendering

	void Rotate(Basis &rotation);											//	apply transformation matrix
	void TranslateZ(float z);												//	translate all points in z direction

	void Render(BView *theView);											//	render the torus in the specified view

	bool TogglePoint(BPoint where, bool shiftIsDown);						//	toggle the existence of a point
	bool TogglePointSelection(BPoint where, bool shiftIsDown);				//	toggle whether a given point is selected
	bool MakePoint(Tile *target);											//	make the point corresponding to the tile
	bool DeleteSelectedPoints();											//	delete all selected points
	bool DeletePoint(Tile *target);											//	delete the specified point
	bool ClearPointSelection();												//	clears the array of selected points - returns whether any were selected

	bool MakeLineFromSelection();											//	makes lines, based on the selection of points
	bool MakeLine(Tile *firstPoint, Tile *secondPoint);						//	makes a line, based on two indices into list of points
	bool DeleteSelectedLines();												//	deletes all selected lines
	bool DeleteLine(long whichLine);										//	deletes the specified line
	bool ClearLineSelection();												//	clears the array of selected lines - returns whether any were selected

	bool PointOnLine(Tile *thePoint, Line *theLine);						//	tests whether a point is on a line

	Tile *FindTile(BPoint where);											//	finds the tile corresponding with a point
	void GetCoords(TilePointer theTile, long &theta, long &phi);			//	given a tile, figure out the theta, phi indices
	TilePointer GetTile(long theta, long phi);								//	given indices, return a pointer to the tile
	bool RotateTiles(long direction);										//	rotates the contents of the tiles
	
	char *MakeText();														//	creates a text version of the torus
	}; // end of Torus class
