//	74.442	Advanced Design & Analysis of Algorithms
//	Spring 1998
//	Dr. Bill Kocay
//
//	Hamish Carr
//	Project:	Toroidal Surface Geometries
//
//

#include "Tile.h"

Tile::Tile()																//	constructor
	{		
	hasPoint = false;														//	mark it as pointless
	isSelected = false;														//	and unselected
	segmentList = NULL;														//	make sure the segment list is empty		
	} // end of Tile() constructor

Tile::~Tile()																//	destructor
	{
	SegmentNode *aSeg;														//	a node on the segment list
	while (segmentList != NULL)												//	i.e. walk down the list
		{
		aSeg = segmentList;													//	grab this node
		segmentList = segmentList->nextSegment;								//	step to next segment
		delete aSeg;														//	and get rid of this one
		} // end of loop down list	
	} // end of Tile() constructor

void Tile::PrintToStream()													//	dump contents to output stream
	{
//	printf("Tile (%d, %d)\n", i, j);
	printf("Tile position:       "); position.PrintToStream();					
	printf("Tile normal  :       "); normal.PrintToStream();
	printf("Tile intensity:      %d\n", intensity);
//	printf("Top left corner:     "); topLeft->PrintToStream();
//	printf("Top right corner:    "); topRight->PrintToStream();
//	printf("Bottom left corner:  "); bottomLeft->PrintToStream();
//	printf("Bottom right corner: "); bottomRight->PrintToStream(); 
	} // end of PrintToStream()

bool Tile::Contains(BPoint P)												//	tests whether a point is inside the tile
	{
	BPoint PtL, PtR, PbL, PbR;												//	pointers from P to the corners
	PtL = (*topLeft) - P;													//	calculate them
	PtR = (*topRight) - P;
	PbL = (*bottomLeft) - P;
	PbR = (*bottomRight) - P;

	if (intensity > 0)														//	if intensity > 0, we want all crossproducts +ve
		{
		if ((PtL.x*PtR.y - PtL.y*PtR.x) > 0)								//	if the angle is > 180
			return false;	
		if ((PtR.x*PbR.y - PtR.y*PbR.x) > 0)								//	if the angle is > 180
			return false;	
		if ((PbR.x*PbL.y - PbR.y*PbL.x) > 0)								//	if the angle is > 180
			return false;	
		if ((PbL.x*PtL.y - PbL.y*PtL.x) > 0)								//	if the angle is > 180
			return false;	
		return true;														//	they all succeed, so return true
		} // end of case of positive intensity
	else if (intensity < 0)													//	we ignore intensity = 0
		{
		if ((PtL.x*PtR.y - PtL.y*PtR.x) < 0)								//	if the angle is > 180
			return false;	
		if ((PtR.x*PbR.y - PtR.y*PbR.x) < 0)								//	if the angle is > 180
			return false;	
		if ((PbR.x*PbL.y - PbR.y*PbL.x) < 0)								//	if the angle is > 180
			return false;	
		if ((PbL.x*PtL.y - PbL.y*PtL.x) < 0)								//	if the angle is > 180
			return false;	
		return true;														//	they all succeed, so return true
		} // end of case for negative intensity
	else return false;														//	play safe, return false for intensity = 0
	} // end of Contains()

void Tile::AddSegment(Segment *newSegment)									//	adds a segment to the segment list
	{
	SegmentNode *oldList = segmentList;										//	temporary pointer to new node
	
	segmentList = new SegmentNode;											//	create a new segment node
	segmentList->nextSegment = oldList;										//	attach the existing list to it
	segmentList->theSegment = newSegment;									//	and attach the segment to it
	} // end of AddSegment()

void Tile::RemoveSegment(Segment *oldSegment)								//	removes a segment from the segment list
																			//	assumes the segment is on the list
	{
	SegmentNode *walker;													//	pointer to walk down list
	SegmentNode *delSeg;													//	pointer for deletion
	
	if (segmentList->theSegment == oldSegment)								//	if the segment is at the top of the list
		{
		delSeg = segmentList;												//	pick up the top node
		segmentList = segmentList->nextSegment;								//	point to the rest of the list
		delete delSeg;														//	delete the top node
		return;																//	we're done, so bail out
		} // end of case where segment at top
	
	for (walker = segmentList; walker->nextSegment != NULL; walker = walker->nextSegment)
		{																	//	walk down the list
		if (walker->nextSegment->theSegment == oldSegment)					//	if the next segment is the one
			{
			delSeg = walker->nextSegment;									//	pick up the node to delete
			walker->nextSegment = delSeg->nextSegment;						//	point past it
			delete delSeg;													//	delete the segment
			return;															//	we're done, so leave			
			} // end of case where we find it
		} // end of case where segment not at top
	} // end of RemoveSegment()

int compareTiles(const void *e1, const void *e2)							//	routine for comparing tiles
	{
	int value = (**((Tile **)e1)).position.z - (**((Tile **)e2)).position.z;//	compare the z coordinates
	return value;															//	sorts into forward order
	} // end of compareTiles
	
	