//	74.442	Advanced Design & Analysis of Algorithms
//	Spring 1998
//	Dr. Bill Kocay
//
//	Hamish Carr
//	Project:	Toroidal Surface Geometries
//
//

#include "Torus.h"

Torus::Torus(float RingRadius, float Radius, long NPhiDivisions, long NThetaDivisions)	
																			//	basic constructor
	{
	itsLocation = NULL;														//	all toruses are anonymous to start with
	name = "Untitled";														//	and untitled
	
	theGridVectors = NULL;													//	set all arrays to null
	theGridPoints = NULL;													//	the grid points (after projection)
	theTiles = NULL;														//	array with all the tiles
	sortedTiles = NULL;														//	array of pointers to the tiles in sorted order

	ConstructTiles(RingRadius, Radius, NPhiDivisions, NThetaDivisions);		//	construct the tiles
	} // end of basic constructor

Torus::Torus(char *torusText, BMessage *newLocation)						//	string constructor
	{
	CharBuffer readBuffer(torusText);										//	create a buffer, and load the text
	char *lineOfText = NULL;												//	line of text
	long i;																	//	loop index
	long theta, phi, dTheta, dPhi;											//	for reading in points and lines
	CharBuffer *lineBuffer;													//	buffer for processing a line
	char *errMsg = new char[100];											//	error message to throw
	Tile *tile1, *tile2;													//	pointers to tiles for constructing lines	
	long nPts, nLns;														//	may be different from actual numbers	
		
	itsLocation = newLocation;												//	set the location

	theGridVectors = NULL;													//	set all arrays to null
	theGridPoints = NULL;													//	the grid points (after projection)
	theTiles = NULL;														//	array with all the tiles
	sortedTiles = NULL;														//	array of pointers to the tiles in sorted order
	
	try 
		{
		lineOfText = readBuffer.ReadLine();									//	retrieve a line
		if (strcmp(lineOfText, torusTitle) != 0)							//	if it isn't what it should be
			throw("Title line missing");									//	throw an error message
		delete lineOfText;													//	delete the line

		name = readBuffer.ReadLine();										//	retrieve the name of the torus			
		if (name == NULL) throw("Name missing");							//	throw an error message if needed

		lineOfText = readBuffer.ReadLine();									//	read the next line
		if (strcmp(lineOfText, tubeTitle) != 0)								//	if it isn't what it should be
			throw("Tube radius title line missing");						//	throw an error message
		delete lineOfText;													//	delete the line

		lineOfText = readBuffer.ReadLine();									//	read the next line
		lineBuffer = new CharBuffer(lineOfText);							//	create a line buffer
		radius = lineBuffer->GetFloat();									//	retrieve the tube radius
		delete lineBuffer;													//	get rid of the line buffer

		lineOfText = readBuffer.ReadLine();									//	read the next line
		if (strcmp(lineOfText, ringTitle) != 0)								//	if it isn't what it should be
			throw("Ring radius title line missing");						//	throw an error message
		delete lineOfText;													//	delete the line

		lineOfText = readBuffer.ReadLine();									//	read the next line
		lineBuffer = new CharBuffer(lineOfText);							//	create a line buffer
		ringRadius = lineBuffer->GetFloat();								//	retrieve the ring radius
		delete lineBuffer;													//	get rid of the line buffer

		lineOfText = readBuffer.ReadLine();									//	read the next line
		if (strcmp(lineOfText, nPhiTitle) != 0)								//	if it isn't what it should be
			throw("N Phi Divisions title line missing");					//	throw an error message
		delete lineOfText;													//	delete the line

		lineOfText = readBuffer.ReadLine();									//	read the next line
		lineBuffer = new CharBuffer(lineOfText);							//	create a line buffer
		nPhiDivisions = lineBuffer->GetLong();								//	retrieve the # of phi divs.
		delete lineBuffer;													//	get rid of the line buffer

		lineOfText = readBuffer.ReadLine();									//	read the next line
		if (strcmp(lineOfText, nThetaTitle) != 0)							//	if it isn't what it should be
			throw("N Theta Divisions title line missing");					//	throw an error message
		delete lineOfText;													//	delete the line

		lineOfText = readBuffer.ReadLine();									//	read the next line
		lineBuffer = new CharBuffer(lineOfText);							//	create a line buffer
		nThetaDivisions = lineBuffer->GetLong();							//	retrieve the # of theta divs.
		delete lineBuffer;													//	get rid of the line buffer

		ConstructTiles(ringRadius, radius, nPhiDivisions, nThetaDivisions);	//	construct the tiles

		lineOfText = readBuffer.ReadLine();									//	read the next line
		if (strcmp(lineOfText, nPointsTitle) != 0)							//	if it isn't what it should be
			throw("N Points title line missing");							//	throw an error message
		delete lineOfText;													//	delete the line

		lineOfText = readBuffer.ReadLine();									//	read the next line
		lineBuffer = new CharBuffer(lineOfText);							//	create a line buffer
		nPts = lineBuffer->GetLong();										//	retrieve the # of points
		delete lineBuffer;													//	get rid of the line buffer

		for (i = 0; i < nPts; i++)											//	walk through the points individually
			{
			lineOfText = readBuffer.ReadLine();								//	read the next line
			lineBuffer = new CharBuffer(lineOfText);						//	create a line buffer
			
			theta = lineBuffer->GetLong();									//	retrieve the theta coordinate
			if ((theta < 0) || (theta >= nThetaDivisions))					//	if it's an illegal value
				{
				sprintf(errMsg, "Illegal value for theta (%d)", theta);		//	prepare the error message
				throw(errMsg);												//	and throw it									
				} // end of test for invalid theta
			phi = lineBuffer->GetLong();									//	retrieve the phi coordinate
			if ((phi < 0) || (phi >= nPhiDivisions))						//	if it's an illegal value
				{
				sprintf(errMsg, "Illegal value for phi (%d)", phi);			//	prepare the error message
				throw(errMsg);												//	and throw it									
				} // end of test for invalid phi
						
			tile1 = GetTile(theta, phi);									//	retrieve the correct tile
			MakePoint(tile1);												//	add a point to it						
			} // end of loop to read in points
				
		lineOfText = readBuffer.ReadLine();									//	read the next line
		if (strcmp(lineOfText, nLinesTitle) != 0)							//	if it isn't what it should be
			throw("N Lines title line missing");							//	throw an error message
		delete lineOfText;													//	delete the line

		lineOfText = readBuffer.ReadLine();									//	read the next line
		lineBuffer = new CharBuffer(lineOfText);							//	create a line buffer
		nLns = lineBuffer->GetLong();										//	retrieve the # of lines
		delete lineBuffer;													//	get rid of the line buffer

		for (i = 0; i < nLns; i++)											//	walk through the lines individually
			{
			lineOfText = readBuffer.ReadLine();								//	read the next line
			lineBuffer = new CharBuffer(lineOfText);						//	create a line buffer
			
			theta = lineBuffer->GetLong();									//	retrieve the theta coordinate
			if ((theta < 0) || (theta >= nThetaDivisions))					//	if it's an illegal value
				{
				sprintf(errMsg, "Illegal value for theta (%d)", theta);		//	prepare the error message
				throw(errMsg);												//	and throw it									
				} // end of test for invalid theta
			phi = lineBuffer->GetLong();									//	retrieve the phi coordinate
			if ((phi < 0) || (phi >= nPhiDivisions))						//	if it's an illegal value
				{
				sprintf(errMsg, "Illegal value for phi (%d)", phi);			//	prepare the error message
				throw(errMsg);												//	and throw it									
				} // end of test for invalid phi
						
			tile1 = GetTile(theta, phi);									//	retrieve the correct tile

			lineOfText = readBuffer.ReadLine();								//	read the next line
			lineBuffer = new CharBuffer(lineOfText);						//	create a line buffer
			
			dTheta = lineBuffer->GetLong();									//	retrieve the theta coordinate
			if (abs(dTheta*2) > nThetaDivisions)							//	if it's an illegal value
				{
				sprintf(errMsg, "Illegal value for dTheta (%d)", dTheta);	//	prepare the error message
				throw(errMsg);												//	and throw it									
				} // end of test for invalid theta
			dPhi = lineBuffer->GetLong();									//	retrieve the phi coordinate
			if (abs(dPhi*2) > nPhiDivisions)								//	if it's an illegal value
				{
				sprintf(errMsg, "Illegal value for dPhi (%d)", dPhi);		//	prepare the error message
				throw(errMsg);												//	and throw it									
				} // end of test for invalid phi
			
			theta += dTheta;												//	construct the point theta
			if (theta < 0) theta += nThetaDivisions;						//	trim at low end
			if (theta >= nThetaDivisions) theta -= nThetaDivisions;			//	and chop at high end

			phi += dPhi;													//	construct the point phi
			if (phi < 0) phi += nPhiDivisions;								//	trim at low end
			if (phi >= nPhiDivisions) phi -= nPhiDivisions;					//	and chop at high end

			tile2 = GetTile(theta, phi);									//	retrieve the tile
			
			MakeLine(tile1, tile2);											//	construct the line

			} // end of loop to read in lines
		ClearPointSelection();												//	clear the selection
		} // end of try block
	catch (CharBuffer::OutOfMemory *e)										//	if there is a memory error
		{
		DoAlert("Ran out of memory trying to allocate torus.\n\nCreating default torus.\n");		
		ConstructTiles(2000, 1000, 24, 120);								//	construct a default torus		
		} // end of out of memory error
	catch (CharBuffer::EndOfBufferHit *e)									//	if the file was faulty
		{
		DoAlert("Ran out of data trying to read torus in.\n\nCreating default torus.\n");
		ConstructTiles(2000, 1000, 24, 120);								//	construct a default torus		
		} // end of out of data error
	catch (char *errorMessage)												//	catch those little ones
		{
		DoAlert(errorMessage);												//	do the alert
		ConstructTiles(2000, 1000, 24, 120);								//	construct a default torus		
		} // end of out of data error
	delete errMsg;															//	delete the buffer
	} // end of string constructor

void Torus::ConstructTiles(float RingRadius, float Radius, long NPhiDivisions, long NThetaDivisions)
																			//	constructs the tiles for the torus
	{
	long i, j;																//	loop indices
	float theta, phi;														//	angles
	float d_theta, d_phi;													//	angle increments
	float *rsinPhi = new float[2*NPhiDivisions];							//	local tables for sin, cosine
	float *RrcosPhi = new float[2*NPhiDivisions];							//	doubled b/c we need midpoints as well
	float *sinTheta = new float[2*NThetaDivisions];
	float *cosTheta = new float[2*NThetaDivisions];
	float *sinPhi = new float[2*NPhiDivisions];
	float *cosPhi = new float[2*NPhiDivisions];
	
	radius = Radius; 														//	just copy them each in
	ringRadius = RingRadius; 
	nPhiDivisions = NPhiDivisions;
	nThetaDivisions = NThetaDivisions;
	nTiles = nPhiDivisions*nThetaDivisions;

	if (theGridVectors != NULL) delete theGridVectors;						//	clear any existing arrays
	if (theGridPoints != NULL) delete theGridPoints;						
	if (theTiles != NULL) delete theTiles;						
	if (sortedTiles != NULL) delete sortedTiles;						

	Tile *theTile = theTiles = new Tile[nTiles];							//	tile being filled in
	Vector *theVector = theGridVectors = new Vector[nTiles];		
	BPoint *thePoint = theGridPoints = new BPoint[nTiles];		
	TilePointer *theSorter = sortedTiles = new TilePointer[nTiles];

	torusOrigin = Vector(0, 0, 0);											//	set the origin of the torus
	torusBasis = Basis(Vector(1, 0, 0), Vector(0, 1, 0), Vector(0, 0, 1));	//	and the basis for its coordinates

	theCurrentTool = selectionTool;											//	set the initial tool
	nLines = 0; nPoints = 0;												//	set the initial # of points & lines
	nPointsSelected = 0;													//	and the number of selected points	
	nLinesSelected = 0;														//	not to mention the number of selected lines	
		
	theta = 0; 																//	start our angle off right
	d_theta = PI / nThetaDivisions;											//	find d_theta
	phi = 0;																//	do the same for phi
	d_phi = PI / nPhiDivisions;												//	find d_phi
	
	for (i = 0; i < 2*nThetaDivisions; i++, theta += d_theta)				//	walk through theta arrays
		{
		sinTheta[i] = sin(theta);											//	calculate the values
		cosTheta[i] = cos(theta);										
		} // end of loop through theta arrays	
	
	for (j = 0; j < 2*nPhiDivisions; j++, phi += d_phi)						//	walk through phi arrays
																			//	we don't need all of it, but it's easier
		{
		sinPhi[j] = sin(phi);												//	fill in each item
		rsinPhi[j] = radius*sinPhi[j];
		cosPhi[j] = cos(phi);	
		RrcosPhi[j] = ringRadius + radius*cosPhi[j];							
		} // end of loop through phi arrays
	
	for (i = 0; i < 2*nThetaDivisions; i += 2)								//	walk around the ring
		for (j = 0; j < 2*nPhiDivisions;)									//	walk around the tube
			{
			theVector->x = cosTheta[i]*RrcosPhi[j];							//	fill in the components of the vector
			theVector->y = sinTheta[i]*RrcosPhi[j];							//	move it over
			theVector->z = rsinPhi[j];
			
			j++;															//	step up to the midpoint of a tile
			
			theTile->position.x = cosTheta[i+1]*RrcosPhi[j];				//	set the location of the tile
			theTile->position.y = sinTheta[i+1]*RrcosPhi[j];				
			theTile->position.z = rsinPhi[j];				
			theTile->normal.x = cosTheta[i+1]*cosPhi[j];					//	calculate the normal vector
			theTile->normal.y = sinTheta[i+1]*cosPhi[j];
			theTile->normal.z = sinPhi[j];
			theTile->topLeft = thePoint;									//	set the top left point
			if (i == 2*nThetaDivisions - 2)									//	if we are in the last ring
				theTile->topRight = thePoint + nPhiDivisions - nTiles;
			else
				theTile->topRight = thePoint + nPhiDivisions;
			if (j == 2*nPhiDivisions - 1)									//	if we are at the end of the ring
				{
				theTile->bottomLeft = thePoint + 1 - nPhiDivisions;			//  point to the start of the ring
				theTile->bottomRight = theTile->topRight + 1 - nPhiDivisions;
				} // end of case at end of ring
			else // i.e. somewhere in middle
				{
				theTile->bottomLeft = thePoint + 1;							// point to next tile
				theTile->bottomRight = theTile->topRight + 1;
				} // end of other case (middle of ring)			
	
			*theSorter = theTile;											//	put it in the sorting array
			
			j++;															//	increment our indices
			theVector++; theTile++; thePoint++; theSorter++; 				//	and our pointers
			} // end of loop around tube
		
	for (i = 0; i < maxNPoints; i++)										//	clear out the arrays of clicked points
		{
		thePoints[i] = NULL;												//	set them to have NULL tiles
		selectedPoints[i] = NULL;
		} // end of loop to clear out point arrays

	delete rsinPhi;															//	dispose of dynamic memory
	delete RrcosPhi;
	delete sinTheta;
	delete cosTheta;
	delete sinPhi;
	delete cosPhi;
	} // end of ConstructTiles()
	
void Torus::SetRenderingParms(BPoint ScreenCentre, float ViewDistance, float ViewScale)
																			//	sets the variables for rendering
	{
	screenCentre = ScreenCentre;											//	copy the screen centre in as well
	viewDistance = ViewDistance;											//	and the view distance
	viewScale = ViewScale;													//	hell - may as well copy the scale
	} // end of SetRenderingParms()

Torus::~Torus()																//	short & sweet destructor
	{
	delete theGridVectors;													//	delete everything we created
	delete theGridPoints;
	delete sortedTiles;
	delete theTiles;	
	} // end of Torus destructor
	
void Torus::Render(BView *theView)											//	render the torus in the specified view
	{
	short intense;															//	local temporary
	Tile **walker, **stop;													//	for walking through sorted list
	BPoint *tL, *tR, *bL, *bR;												//	local copies
	BPoint centre;															//	centre of the tile
	register Tile *walkptr;													//	dereferenced copy of walker
	BPicture *thePicture = NULL;											//	recorded picture calls
	SegmentNode *whichNode;													//	pointer for walking down list of segments
	Segment *whichSeg;														//	pointer to the segment proper
	BPoint inPoint, outPoint;												//	endpoints of a segment on the tile
	BPoint deltaPoint;														//	difference between endpoints
//	char numBuf[32];														//	buffer for testing only
//	long i, j;																	//	for testing only

	theView->BeginPicture(new BPicture);									//	start capturing calls
//	theView->SetFontSize(9);													//	set the font size
	theView->SetHighColor(0, 0, 0, 0);										//	set it to black
	theView->FillRect(theView->Bounds(), B_SOLID_HIGH);						//	start off setting everything to black
	stop = sortedTiles+nTiles;												//	set our sentinel pointer
	for (walker = stop - 1; walker >= sortedTiles; walker--)				//	walk along the sorted array
		{
		walkptr = *walker;													//	grab the pointer											
		if (walkptr->position.z <= 0)										//	i.e. it (and all later ones) is behind us
			break;
		intense = walkptr->intensity;										//	retrieve the intensity
		if (intense != 0)													//	don't paint it if it is black
			{
			if (insideTorus && intense > 0)									//	if the point is inside the torus & the face is facing away
				theView->SetHighColor(intense, 0, 0, 0);					//	set it to a shade of red
			else
				theView->SetHighColor(0, abs(intense), 0, 0);				//	set the colour
			tL = walkptr->topLeft; tR = walkptr->topRight;					//	make local copies
			bL = walkptr->bottomLeft; bR = walkptr->bottomRight;			//	make local copies
			theView->FillTriangle(*tL, *tR, *bR);							//	draw the first triangle
			theView->FillTriangle(*tL, *bR, *bL);							//	and the second
			if ((**walker).hasPoint)										//	if there is a point in the middle of the face
				{
				theView->SetHighColor(0, 0, 0, 0);							//	set the colour to black
				centre = *tL + *tR + *bL + *bR;								//	add the 4 points together
				centre.x /= 4; centre.y /= 4;								//	and divide to get the average
				theView->FillEllipse(centre, 3, 3);							//	draw a point
				if ((**walker).isSelected)									//	if the point is marked as selected
					theView->StrokeEllipse(centre, 5, 5);					//	draw a circle around it
				} // end of test for point
			for (whichNode = (**walker).segmentList; whichNode != NULL; whichNode = whichNode->nextSegment)
																			//	walk down the list of segments
				{
				whichSeg = whichNode->theSegment;							//	pick up the segment
				
				inPoint = BPoint(0,0);										//	start the sum off
				AddScaledPoint(inPoint, tL, whichSeg->inTL);				//	and add all the points to it
				AddScaledPoint(inPoint, tR, whichSeg->inTR);	
				AddScaledPoint(inPoint, bR, whichSeg->inBR);	
				AddScaledPoint(inPoint, bL, whichSeg->inBL);					

				outPoint = BPoint(0,0);										//	start the sum off
				AddScaledPoint(outPoint, tL, whichSeg->outTL);				//	and add all the points to it
				AddScaledPoint(outPoint, tR, whichSeg->outTR);	
				AddScaledPoint(outPoint, bR, whichSeg->outBR);	
				AddScaledPoint(outPoint, bL, whichSeg->outBL);	
		
				theView->SetHighColor(0, 0, 0, 0);							//	set the colour to black
				theView->StrokeLine(inPoint, outPoint);						//	connect them with a line				
				} // end of test for segments
			// this chunk for testing only
//			theView->SetHighColor(0, 0, 0, 0);								//	set the colour to black
//			centre = *tL + *tR + *bL + *bR;									//	add the 4 points together
//			centre.x /= 4; centre.y /= 4;									//	and divide to get the average
//			centre.x -= 6; centre.y -= 4;									//	move it backwards & upwards 
//			theView->MovePenTo(centre.x, centre.y);							//	move to that place
//			GetCoords(*walker, i, j);										//	retrieve coords of the tile
//			sprintf(numBuf, "(%d, %d)", i, j);								//	put the coords in a string
//			theView->DrawString(numBuf);									//	draw the coords on screen
//			theView->StrokeTriangle(*((**walker).topLeft), *((**walker).topRight), *((**walker).bottomRight));
//			theView->StrokeTriangle(*((**walker).topLeft), *((**walker).bottomRight), *((**walker).bottomLeft));
//			theView->Sync();
			} // end of test for intensity
		} // end of loop through array
	theView->SetHighColor(0, 0, 0, 0);										//	set the colour back to black
	thePicture = theView->EndPicture();										//	stop capturing calls
	theView->MovePenTo(0,0);												//	make sure we are in the right place
	theView->DrawPicture(thePicture);										//	and draw the picture properly
	delete thePicture;														//	get rid of the picture
	} // end of Render()

void Torus::GetCoords(TilePointer theTile, long &theta, long &phi)			//	given a tile, figure out the theta, phi indices
	{
	long nTiles;															//	distance in tiles (through array)
	
	nTiles = theTile - theTiles;											//	compute the offset from the front of the array
	theta = nTiles / nPhiDivisions;											//	compute theta
	phi = nTiles % nPhiDivisions;											//	and phi	
	} // end of GetCoords()
	
TilePointer Torus::GetTile(long theta, long phi)							//	given indices, return a pointer to the tile
	{
	return theTiles+(theta*nPhiDivisions)+phi;								//	return the appropriate index
	} // end of GetTile()
	
char *Torus::MakeText()														//	creates a text version of the torus
	{
	CharBuffer theBuffer(16384);											//	set up a char buffer with a biggish string size									
	char stringBuf[256];													//	local buffer for constructing lines
	long i;																	//	loop index
	long theta, phi;														//	theta, phi indices for points

	try
		{
		sprintf(stringBuf,"%s\n%s\n", torusTitle, name);					//	set up the first two lines
		theBuffer.Write(stringBuf);											//	copy it to the main buffer
		sprintf(stringBuf, "%s\n%f\n", tubeTitle, radius);					//	print out the radius
		theBuffer.Write(stringBuf);											//	copy it to the main buffer
		sprintf(stringBuf, "%s\n%f\n", ringTitle, ringRadius);				//	and the radius of the ring
		theBuffer.Write(stringBuf);											//	copy it to the main buffer
		sprintf(stringBuf, "%s\n%d\n", nPhiTitle, nPhiDivisions);			//	print out the # of phi divisions
		theBuffer.Write(stringBuf);											//	copy it to the main buffer
		sprintf(stringBuf, "%s\n%d\n", nThetaTitle, nThetaDivisions);		//	print out the # of phi divisions
		theBuffer.Write(stringBuf);											//	copy it to the main buffer
		sprintf(stringBuf, "%s\n%d\n", nPointsTitle, nPoints);				//	print out the # of points
		theBuffer.Write(stringBuf);											//	copy it to the main buffer
		for (i = 0; i < nPoints; i++)										//	loop through all the points	
			{
			GetCoords(thePoints[i], theta, phi);							//	retrieve the coordinates of each
			sprintf(stringBuf, "%d %d\n", theta, phi);						//	print out the coordinates
			theBuffer.Write(stringBuf);										//	copy it to the main buffer
			} // end of loop through points
		sprintf(stringBuf, "%s\n%d\n", nLinesTitle, nLines);				//	print out the # of points
		theBuffer.Write(stringBuf);											//	copy it to the main buffer
		for (i = 0; i < nLines; i++)										//	loop through all the lines	
			{
			sprintf(stringBuf, "%d %d\n", theLines[i].theta, theLines[i].phi);		
																			//	print out the coordinates
			theBuffer.Write(stringBuf);										//	copy it to the main buffer
			sprintf(stringBuf, "%d %d\n", theLines[i].dTheta, theLines[i].dPhi);		
																			//	print out the coordinates
			theBuffer.Write(stringBuf);										//	copy it to the main buffer
			} // end of loop through lines
		return theBuffer.GetContents();										//	return whatever the buffer has in it
		} // end of try block
	catch (CharBuffer::OutOfMemory *e)										//	and if it throws an exception
		{	
		return NULL;														//	return a NULL to indicate failure
		} // end of catch block
	} // end of MakeText()
	